import unittest
import big_o
from big_o import reports

class TestReport(unittest.TestCase):

    def test_report_return(self):
        best, others = big_o.big_o(sorted, lambda n: big_o.datagen.integers(n, 100, 500))

        assert isinstance(reports.big_o_report(best, others), str)

    def test_report_return_raw_data_true(self):
        best, others = big_o.big_o(sorted, lambda n: big_o.datagen.integers(n, 100, 500), return_raw_data=True)

        report = reports.big_o_report(best, others)
        assert 'measures' not in report
        assert 'times' not in report




from __future__ import absolute_import

from timeit import Timer

import numpy as np

from big_o.complexities import ALL_CLASSES


def measure_execution_time(func, data_generator,
                           min_n=100, max_n=100000, n_measures=10,
                           n_repeats=1, n_timings=1):
    """ Measure the execution time of a function for increasing N.

    Input:
    ------

    func -- Function of which the execution time is measured.
            The function is called as func(data), where data is returned
            by the argument `data_generator`

    data_generator -- Function returning input data of 'length' N.
                      Input data for the argument `func` is created as
                      `data_generator(N)`. Common data generators are defined
                      in the submodule `big_o.datagen`

    min_n, max_n, n_measures -- The execution time of func is measured
                                at `n_measures` points between `min_n` and
                                `max_n` (included)

    n_repeats -- Number of times func is called to compute execution time
                 (return the cumulative time of execution)

    n_timings -- Number of times the timing measurement is repeated.
                 The minimum time for all the measurements is kept.

    Output:
    -------

    n -- List of N's used as input to `data_generator`

    time -- List of total execution time for each N in seconds
    """

    # we need a wrapper that holds a reference to func and the generated data
    # for the timeit.Timer object
    class func_wrapper(object):

        def __init__(self, n):
            self.data = data_generator(n)

        def __call__(self):
            return func(self.data)

    # TODO: check that max_n is not larger than max int64
    ns = np.linspace(min_n, max_n, n_measures).astype('int64')
    execution_time = np.empty(n_measures)
    for i, n in enumerate(ns):
        timer = Timer(func_wrapper(n))
        measurements = timer.repeat(n_timings, n_repeats)
        execution_time[i] = np.min(measurements)
    return ns, execution_time


def infer_big_o_class(ns, time, classes=ALL_CLASSES, verbose=False, simplicity_bias=1e-6):
    """Infer the complexity class from execution times.

    Input:
    ------

    ns -- Array of values of N for which execution time has been measured.

    time -- Array of execution times for each N in seconds.

    classes -- The complexity classes to consider. This is a list of subclasses
               of `big_o.complexities.ComplexityClass`.
               Default: all the classes in `big_o.complexities.ALL_CLASSES`

    verbose -- If True, print parameters and residuals of the fit for each
               complexity class

    simplicity_bias -- Preference toward choosing simpler methods when
                       the difference between residuals is less than the
                       simplicity_bias. If simplicity_bias is 0, the
                       complexity class with the lowest residuals is
                       always chosen.

    Output:
    -------

    best_class -- Object representing the complexity class that best fits
                  the measured execution times.
                  Instance of `big_o.complexities.ComplexityClass`.

    fitted -- A dictionary of fittest complexity classes to the fit residuals
    """

    best_class = None
    best_residuals = np.inf
    fitted = {}
    for class_ in classes:
        inst = class_()
        residuals = inst.fit(ns, time)
        fitted[inst] = residuals

        # NOTE: subtract bias for tiny preference for simpler methods
        # TODO: improve simplicity bias (AIC/BIC)?
        if residuals < best_residuals - simplicity_bias:
            best_residuals = residuals
            best_class = inst
        if verbose:
            print(inst, '(r={:f})'.format(residuals))
    return best_class, fitted


def big_o(func, data_generator,
          min_n=100, max_n=100000, n_measures=10,
          n_repeats=1, n_timings=1, classes=ALL_CLASSES, verbose=False, return_raw_data=False):
    """ Estimate time complexity class of a function from execution time.

    Input:
    ------

    func -- Function of which the execution time is measured.
            The function is called as func(data), where data is returned
            by the argument `data_generator`

    data_generator -- Function returning input data of 'length' N.
                      Input data for the argument `func` is created as
                      `data_generator(N)`. Common data generators are defined
                      in the submodule `big_o.datagen`

    min_n, max_n, n_measures -- The execution time of func is measured
                                at `n_measures` points between `min_n` and
                                `max_n` (included)

    n_repeats -- Number of times func is called to compute execution time
                 (return the cumulative time of execution)

    n_timings -- Number of times the timing measurement is repeated.
                 The minimum time for all the measurements is kept.

    classes -- The complexity classes to consider. This is a list of subclasses
               of `big_o.complexities.ComplexityClass`.
               Default: all the classes in `big_o.complexities.ALL_CLASSES`

    verbose -- If True, print parameters and residuals of the fit for each
               complexity class

    return_raw_data -- If True, the function returns the measure points and its
                       corresponding execution times as part of the fitted dictionary
                       of complexity classes. When this flag is true, fitted will
                       contain the entries:
                       {... 'measures': [<int>+], 'times': [<float>+] ...}

    Output:
    -------

    best_class -- Object representing the complexity class that best fits
                  the measured execution times.
                  Instance of `big_o.complexities.ComplexityClass`.

    fitted -- A dictionary of fittest complexity classes to the fit residuals
    """

    ns, time = measure_execution_time(func, data_generator,
                                      min_n, max_n, n_measures, n_repeats,
                                      n_timings)
    best, fitted = infer_big_o_class(ns, time, classes, verbose=verbose)

    if return_raw_data:
        fitted['measures'] = ns
        fitted['times'] = time

    return best, fitted



"""Definition of complexity classes."""

import numpy as np


class NotFittedError(Exception):
    pass


class ComplexityClass(object):
    """ Abstract class that fits complexity classes to timing data.
    """

    #: bool: _recalculate_fit_residuals controls if the residuals value
    # returned from np.linalg.lstsq() is equivalent to the margin of
    # error of the found coefficients.
    #
    # This is normally only needed if the complexity class overrides
    # _transform_time() or _inverse_transform_time()
    _recalculate_fit_residuals = False

    def __init__(self):
        # list of parameters of the fitted function class as returned by the
        # last square method np.linalg.lstsq
        self.coeff = None

    def fit(self, n, t):
        """ Fit complexity class parameters to timing data.

        Input:
        ------

        n -- Array of values of N for which execution time has been measured.

        t -- Array of execution times for each N in seconds.

        Output:
        -------

        residuals -- Sum of square errors of fit
        """
        n = np.asanyarray(n)
        t = np.asanyarray(t)

        x = self._transform_n(n)
        y = self._transform_time(t)
        coeff, residuals, rank, s = np.linalg.lstsq(x, y, rcond=-1)
        self.coeff = coeff

        # Check if residuals from least square can be used, or if it
        # must be explicitly calculated.
        if self._recalculate_fit_residuals:
            ref_t = self.compute(n)
            residuals = np.sum((ref_t - t) ** 2)
        else:
            residuals = residuals[0]
        return residuals

    def compute(self, n):
        """ Compute the value of the fitted function at `n`. """
        if self.coeff is None:
            raise NotFittedError()

        # Result is linear combination of the terms with the fitted
        # coefficients
        x = self._transform_n(n)
        tot = 0
        for i in range(len(self.coeff)):
            tot += self.coeff[i] * x[:, i]
        return self._inverse_transform_time(tot)

    def coefficients(self):
        """ Return coefficients in standard form. """
        if self.coeff is None:
            raise NotFittedError()
        return self.coeff

    def __str__(self):
        prefix = '{}: '.format(self.__class__.__name__)

        if self.coeff is None:
            return prefix + 'not yet fitted'
        return prefix + self.format_str().format(
            *self.coefficients()) + ' (sec)'

    # --- abstract methods

    @classmethod
    def format_str(cls):
        """ Return a string describing the fitted function.

        The string must contain one formatting argument for each coefficient.
        """
        return 'FORMAT STRING NOT DEFINED'

    def _transform_n(self, n):
        """ Terms of the linear combination defining the complexity class.

        Output format: number of Ns x number of coefficients .
        """
        raise NotImplementedError()

    def _transform_time(self, t):
        """ Transform time as needed for fitting.
        (e.g., t->log(t)) for exponential class.
        """
        return t

    def _inverse_transform_time(self, t):
        """ Inverse transform time as needed for compute.
        (e.g., t->exp(t)) for exponential class.
        """
        return t

    def __gt__(self, other):
        return self.order > other.order

    def __lt__(self, other):
        return self.order < other.order

    def __le__(self, other):
        return (self < other) or self == other

    def __ge__(self, other):
        return (self > other) or self == other

    def __eq__(self, other):
        return self.order == other.order

    def __hash__(self):
        return id(self)

# --- Concrete implementations of the most popular complexity classes


class Constant(ComplexityClass):
    order = 10

    def _transform_n(self, n):
        return np.ones((len(n), 1))

    @classmethod
    def format_str(cls):
        return 'time = {:.2G}'


class Linear(ComplexityClass):
    order = 30

    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), n]).T

    @classmethod
    def format_str(cls):
        return 'time = {:.2G} + {:.2G}*n'


class Quadratic(ComplexityClass):
    order = 50

    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), n * n]).T

    @classmethod
    def format_str(cls):
        return 'time = {:.2G} + {:.2G}*n^2'


class Cubic(ComplexityClass):
    order = 60

    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), n ** 3]).T

    @classmethod
    def format_str(cls):
        return 'time = {:.2G} + {:.2G}*n^3'


class Logarithmic(ComplexityClass):
    order = 20

    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), np.log(n)]).T

    @classmethod
    def format_str(cls):
        return 'time = {:.2G} + {:.2G}*log(n)'


class Linearithmic(ComplexityClass):
    order = 40

    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), n * np.log(n)]).T

    @classmethod
    def format_str(cls):
        return 'time = {:.2G} + {:.2G}*n*log(n)'


class Polynomial(ComplexityClass):
    order = 70

    _recalculate_fit_residuals = True

    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), np.log(n)]).T

    def _transform_time(self, t):
        return np.log(t)

    def _inverse_transform_time(self, t):
        return np.exp(t)

    @classmethod
    def format_str(cls):
        return 'time = {:.2G} * x^{:.2G}'

    def coefficients(self):
        """ Return coefficients in standard form. """
        # The polynomial is stored in the format
        # exp(a)*n^b where [a, b] are the coefficients
        # Technical full format is exp(a+b*ln(n))
        #
        # Standard form is a*n^b
        if self.coeff is None:
            raise NotFittedError()

        a, b = self.coeff
        return np.exp(a), b


class Exponential(ComplexityClass):
    order = 80

    _recalculate_fit_residuals = True

    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), n]).T

    def _transform_time(self, t):
        return np.log(t)

    def _inverse_transform_time(self, t):
        return np.exp(t)

    @classmethod
    def format_str(cls):
        return 'time = {:.2G} * {:.2G}^n'

    def coefficients(self):
        """ Return coefficients in standard form. """
        # The polynomial is stored in the format
        # exp(a)*exp(b)^n where [a, b] are the coefficients
        # Technical full format is exp(a+b*n)
        #
        # Standard form is a*b^n
        if self.coeff is None:
            raise NotFittedError()

        a, b = self.coeff
        return np.exp(a), np.exp(b)


ALL_CLASSES = [Constant, Linear, Quadratic, Cubic, Polynomial,
               Logarithmic, Linearithmic, Exponential]



"""Data generators for big_O module."""

import random
import string


def n_(n):
    """ Return N. """
    return n


def range_n(n, start=0):
    """ Return the sequence [start, start+1, ..., start+N-1]. """
    return list(range(start, start+n))


def integers(n, min_, max_):
    """ Return sequence of N random integers between min_ and max_ (included).
    """
    return [random.randint(min_, max_) for _ in range(n)]


def large_integers(n):
    """ Return sequence of N large random integers. """
    return [random.randint(-50, 50) * 1000000 + random.randint(0, 10000)
            for _ in range(n)]


def strings(n, chars=string.ascii_letters):
    """ Return random string of N characters, sampled at random from `chars`.
    """
    return ''.join([random.choice(chars) for i in range(n)])
